/*
 *  Copyright (C) 2000 Lyrisoft Solutions, Inc.
 *
 *  This library is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU Library General Public License as published
 *  by the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package com.lyrisoft.sql;

import java.util.HashSet;
import java.util.HashMap;
import java.util.List;
import java.util.LinkedList;
import java.util.Collection;
import java.util.Iterator;

import java.sql.*;

/**
 * Constructs a "dependency graph" of a list of tables, and returns
 * them in sorted order, such that any dependencies come before their
 * dependents in the returned list.
 */ 
public class DependencyGraph {
    static boolean DEBUG = false;
    HashMap _nodes;
    HashSet _allowedTables;
    DatabaseMetaData _meta;
    List _sorted;

    public DependencyGraph(String[] tables, DatabaseMetaData meta) 
        throws SQLException
    {
        _meta = meta;
        _nodes = new HashMap();
        _allowedTables = new HashSet();
        for (int i=0; i < tables.length; i++) {
            _allowedTables.add(tables[i].toUpperCase());
        }

        for (int i=0; i < tables.length; i++) {
            addNode(tables[i].toUpperCase());
        }
    }

    /**
     * Get the sorted list of tables as a List
     */
    public synchronized List getSorted() {
        if (_sorted != null) {
            return _sorted;
        }

        DEBUG("sort...");
        List in = new LinkedList(_nodes.values());
        _sorted = new LinkedList();

        while (in.size() > 0) {
            DEBUG("input size = " + in.size());
            for (Iterator i = in.iterator(); i.hasNext(); ) {
                Node n = (Node)i.next();
                boolean depsOk = true;
                for (Iterator deps = n.getDeepDependencies().iterator(); 
                     deps.hasNext(); ) 
                {
                    Node dep = (Node)deps.next();
                    if (!_sorted.contains(dep.getName())) {
                        DEBUG(n + " is missing " + dep + "; continuing");
                        depsOk = false;
                        break;
                    }
                }
                if (depsOk) {
                    DEBUG("Deps ok for " + n + "!!!");
                    _sorted.add(n.getName());
                    i.remove();
                }
            }
        }
        return _sorted;
    }

    Iterator getNodes() {
        return _nodes.values().iterator();
    }

    Node getOrCreateNode(String table) {
        Node n = (Node)_nodes.get(table);
        if (n == null) {
            n = new Node(table);
            DEBUG("creating node: " + table);
            _nodes.put(table, n);
        }
        return n;
    }

    void addNode(String table) throws SQLException {
        DEBUG("addNode: " + table);
        Node n = getOrCreateNode(table);
        
        ResultSet rs = _meta.getImportedKeys(null, null, table);
        while (rs.next()) {
            String importTable = rs.getString("PKTABLE_NAME");
            if (_allowedTables.contains(importTable)) {
                DEBUG("\tadding dep: " + importTable);
                n.addDependency(getOrCreateNode(importTable));
            } else {
                DEBUG("\tskipping dep: " + importTable);
            }
        }
    }

    void DEBUG(String s) {
        if (DEBUG) {
            System.err.println(s);
        }
    }
}
