/*
 *  Copyright (C) 2000 Lyrisoft Solutions, Inc.
 *
 *  This library is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU Library General Public License as published
 *  by the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package com.lyrisoft.sql;

import java.util.List;
import java.util.Vector;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Comparator;
import java.util.Iterator;
import java.util.Arrays;
import java.util.StringTokenizer;
import java.io.PrintStream;
import java.sql.*;

/**
 * Creates a SQL script that will recreate a database by generating
 * CREATE TABLE (...) commands followed by INSERT INTO ( ... ) commands.
 *
 * @author Taso Lyristis (taso@lyrisoft.com)
 * @version $Revision: 1.3 $
 */
public class OracleDump {
    /**
     * Starts a new database dump on the specified PrintStream.
     *
     * @param driver the JDBC driver
     * @param user the db user
     * @param pass the db password
     * @param url the db url
     * @param tables an array of table names to dump
     * @param out the printstream to dump to
     */ 
    public static void doDump(String driver, String user, String pass, String url, 
                              String[] tables, PrintStream out) throws SQLException 
    {
        // install the driver
        try {
            Class.forName(driver);
        }
        catch (Exception e) {
            e.printStackTrace();
            throw new RuntimeException("in Dump.doDump(): " + e.toString());
        }

        Connection conn = null;
        try {
            conn = DriverManager.getConnection(url, user, pass);
            DatabaseMetaData metaData = conn.getMetaData();

            System.err.println("sorting tables by dependencies...");
            DependencyGraph graph = new DependencyGraph(tables, metaData);
            List l = graph.getSorted();
            l.toArray(tables);
            System.err.println("done sorting.");
            
            out.println("/* Generated By the Oracle Dump Program */");
            out.println();
            out.println("/* This script will recreate the following tables:");
            for (int i=0; i < tables.length; i++) {
                out.println("   " + tables[i]);
            }
            out.println(" */");
            out.println();

            // dump out "DELETE" portion
            out.println("/* Delete the old schema */");
            for (int i=0; i < tables.length; i++) {
                dumpDelete(tables[i], out);
            }
            out.println();

            // dump out "CREATE TABLE" and "CREATE INDEX" portions
            out.println("/* Schemas */");
            for (int i=0; i < tables.length; i++) {
                System.err.println("Dumping Schema for " + tables[i]);
                dumpSchema(tables[i], metaData, out);
                System.err.println("Dumping Indexes for " + tables[i]);
                dumpIndexes(tables[i], metaData, out);
                out.println();
            }
            out.println();

            // dump out "INSERT" portion
            out.println("/* Data */");
            for (int i=0; i < tables.length; i++) {
                System.err.println("Dumping Data for " + tables[i]);
                dumpData(tables[i], conn, out);
            }
        }
        // No matter what happened, close the connection when we're finished
        finally {
            if (conn != null) {
                conn.close();
            }
        }
    }

    static void dumpDelete(String table, PrintStream out) {
        out.println("DROP TABLE " + table + " CASCADE CONSTRAINTS;");
    }

    static void dumpIndexes(String table, DatabaseMetaData meta, PrintStream out) 
        throws SQLException 
    {
        table = table.toUpperCase();
        ResultSet rs = meta.getIndexInfo(null, null, table, false, false);
        Indexes indexes = new Indexes();
        while (rs.next()) {
            String column = rs.getString("COLUMN_NAME");
            String name = rs.getString("INDEX_NAME");
            boolean unique = rs.getInt("NON_UNIQUE") == 0;

            if (column != null && !name.startsWith("SYS_")) {
                indexes.add(name, column, unique);
            }
        }
        rs.close();

        for (Iterator i = indexes.getNames(); i.hasNext(); ) {
            String name = (String)i.next();
            out.print("CREATE");
            if (indexes.isUnique(name)) {
                out.print(" UNIQUE");
            }
            out.print(" INDEX " + name + " ON " + table + " (");
            int cnt = 0;
            for (Iterator cols = indexes.getColumns(name); cols.hasNext(); ) {
                if (cnt++ > 0) {
                    out.print(", ");
                }
                out.print(cols.next());
            }
            out.println(");");
        }
    }

    /**
     * Dumps the schema for table in the form of a SQL CREATE statement.
     *
     * @param table name of the table
     * @param meta DatabaseMetaData
     * @param out the PrintStream to write to
     */
    static void dumpSchema(String table, DatabaseMetaData meta, PrintStream out)
        throws SQLException
    {
        int cnt = 0; // a counter that we'll reuse over and over

        table = table.toUpperCase();
        out.println("CREATE TABLE " + table + " (");


        // columns
        ResultSet rs = meta.getColumns("%", "%", table, "%");
        while (rs.next()) {
            if (cnt++ > 0) {
                out.println(",");
            }
            String name = rs.getString("COLUMN_NAME");
            String type = rs.getString("TYPE_NAME");

            out.print("    " + name + " " + type);
            if (type.indexOf("NUMBER") >= 0 || type.indexOf("CHAR") >= 0) {
                int size = rs.getInt("COLUMN_SIZE");
                out.print("(" + size + ")");
            }

            // default
            String def = rs.getString("COLUMN_DEF");
            if (def != null) {
                out.print(" DEFAULT " + def.trim());
            }


            // nullable?
            boolean isNull = "YES".equals(rs.getString("IS_NULLABLE"));
            if (!isNull) {
                out.print(" NOT");
            }
            out.print(" NULL");

        }

        rs.close();
        // foreign keys
        rs = meta.getImportedKeys(null, null, table);
        ForeignKeys fkeys = new ForeignKeys();
        while (rs.next()) {
            String refTable = rs.getString("PKTABLE_NAME");
            String refField = rs.getString("PKCOLUMN_NAME"); 
            String col = rs.getString("FKCOLUMN_NAME");
            int deleteRule = rs.getInt("DELETE_RULE");
            fkeys.add(refTable, refField, col, deleteRule);
        }
        for (Iterator tables = fkeys.getTables(); tables.hasNext(); ) {
            String ftable = (String)tables.next();
            out.println(",");
            out.print("    FOREIGN KEY (");
            cnt = 0;
            for (Iterator cols = fkeys.getLocalColumns(ftable); cols.hasNext(); ) {
                if (cnt++ > 0) {
                    out.print(", ");
                }
                String col = (String)cols.next();
                out.print(col);
            }
            out.print(") REFERENCES " + ftable + " (");
            cnt = 0;
            for (Iterator cols = fkeys.getForeignColumns(ftable); cols.hasNext(); ) {
                if (cnt++ > 0) {
                    out.print(", ");
                }
                String col = (String)cols.next();
                out.print(col);
            }
            out.print(")");
            if (fkeys.getDeleteRule(ftable) == DatabaseMetaData.importedKeyCascade) {
                out.print(" ON DELETE CASCADE");
            }
        }

        //primary keys
        rs.close();
        rs = meta.getPrimaryKeys("%", "%", table);
        cnt = 0;
        while (rs.next()) {
            if (cnt++ == 0) {
                out.println(",");
                out.print("    PRIMARY KEY (");
            } else {
                out.print(", ");
            }
            out.print(rs.getString("COLUMN_NAME"));
        }
        if (cnt > 0) {
            out.print(")");
        }

        out.println();
        out.println(");");
    }

    /**
     * Dumps the contents of a table in the form of a series of SQL INSERT statements
     *
     * @param table the name of the table
     * @param conn an open connection to the db that the table lives in
     * @param out the PrintStream to write to
     */
    static void dumpData(String table, Connection conn, PrintStream out) 
        throws SQLException 
    {
        table = table.toUpperCase();
        String query = "SELECT * FROM " + table;
        Statement stmt = null;
        try {
            stmt = conn.createStatement();
            ResultSet rs = stmt.executeQuery(query);
            ResultSetMetaData meta = rs.getMetaData();
            int cols = meta.getColumnCount();
            while (rs.next()) {
                out.println("INSERT INTO " + table);
                out.print("VALUES (");
                
                for (int i = 1; i <= cols; i++) {
                    if (i > 1) {
                        out.print(", ");
                    }
                    // XXX: incomplete
                    switch (meta.getColumnType(i)) {
                    case Types.INTEGER:
                    case Types.NUMERIC:
                        String val = rs.getString(i);
                        out.print(rs.wasNull() ? "null": val);
                        break;
                    case Types.DATE:
                    case Types.TIME:
                    case Types.TIMESTAMP:
                        Timestamp timestamp = rs.getTimestamp(i);
                        out.print(rs.wasNull() ? "null": Util.toSqlDate(timestamp));
                        break;
                    case Types.CLOB:
                        out.print(quote(Util.clob2String(rs.getClob(i))));
                        break;
                    default: // default is STRING
                        out.print(quote(rs.getString(i)));
                    }
                }
                out.println(");");
            }
            out.println();
        }
        finally {
            if (stmt != null) {
                stmt.close();
            }
        }
    }

    /**
     * Util.quote and newline to chr(10) conversion
     */
    static String quote(String s) {
        if (s == null) {
            return null;
        }

        StringTokenizer st = new StringTokenizer(Util.quote(s), "\r\n", true);
        StringBuffer sb = new StringBuffer();
        while (st.hasMoreTokens()) {
            String tok = st.nextToken();
            if (tok.equals("\r") || tok.equals("\n")) {
                sb.append("'||chr(10)||'");
            } else {
                sb.append(tok);
            }
        }
        return sb.toString();
    }

    static void addToVector(HashMap map, String key, String value) {
        Vector v = (Vector)map.get(key);
        
        if (v == null) {
            v = new Vector(2);
            map.put(key, v);
        }
        v.add(value);
    }

    static class Indexes {
        private HashMap indexes = new HashMap();
        private HashSet uniques = new HashSet();

        public void add(String name, String column, boolean unique) {
            addToVector(indexes, name, column);
            if (unique) {
                uniques.add(name);
            }
        }

        public Iterator getNames() {
            return indexes.keySet().iterator();
        }

        public Iterator getColumns(String name) {
            return ((Vector)indexes.get(name)).iterator();
        }

        public boolean isUnique(String name) {
            return uniques.contains(name);
        }
    }

    static class ForeignKeys {
        private HashMap foreignKeys = new HashMap();
        private HashMap localKeys = new HashMap();
        private HashMap deleteRules = new HashMap();

        public void add(String ftable, String fcolumn, String lcolumn, int deleteRule) {
            String key = ftable.toUpperCase();
            addToVector(foreignKeys, ftable, fcolumn);
            addToVector(localKeys, ftable, lcolumn);

            deleteRules.put(ftable, new Integer(deleteRule));
        }

        public int getDeleteRule(String table) {
            return ((Integer)deleteRules.get(table)).intValue();
        }

        public Iterator getLocalColumns(String table) {
            String key = table.toUpperCase();
            Vector v = (Vector)localKeys.get(key);
            return (v == null) ? (new Vector()).iterator() : v.iterator();
        }

        public Iterator getForeignColumns(String table) {
            String key = table.toUpperCase();
            Vector v = (Vector)foreignKeys.get(key);
            return (v == null) ? (new Vector()).iterator() : v.iterator();
        }

        public Iterator getTables() {
            return foreignKeys.keySet().iterator();
        }
    }

    public static void main(String[] args) {
        if (args.length < 5) {
            System.err.println("Usage: com.lyrisoft.sql.OracleDump [driver] [user] [pass] [url] [table1] <table2>...");
            System.exit(1);
        }
        String driver = args[0];
        String user = args[1];
        String pass = args[2];
        String url = args[3];
        String[] tables = new String[args.length-4];
        for (int i=4; i < args.length; i++) {
            tables[i-4] = args[i];
        }

        try {
            doDump(driver, user, pass, url, tables, System.out);
        }
        catch (Exception e) {
            e.printStackTrace();
        }
    }
}
